import os

import requests
from authlib.integrations.flask_client import OAuth
from dotenv import load_dotenv
from flask import make_response, redirect, request

from nerfw.helpers.errors.user_not_allowed import UserNotAllowed
from nerfw.helpers.login_base import LoginBaseClass
from nerfw.server.token_handler import TokenHandler


class LoginGithub(LoginBaseClass):
    """
    Login to GitHub.
    """

    def __init__(self, oauth: OAuth):
        super().__init__()
        client_id, client_secret = self._get_id_and_secret()
        self.github = oauth.register(
            name="github",
            client_id=client_id,
            client_secret=client_secret,
            access_token_url="https://github.com/login/oauth/access_token",
            access_token_params=None,
            authorize_url="https://github.com/login/oauth/authorize",
            authorize_params=None,
            api_base_url="https://api.github.com/",
            client_kwargs={"scope": "user:email"},
        )
        self.token_handler = TokenHandler()

    def login(self):
        """
        Login to GitHub.

        :return: JWT token
        """

        self.logger.info("Authorizing GitHub login...")
        request_token = request.args.get("code")
        access_token = self._get_access_token(request_token)
        user_data = self._get_user_data(access_token)

        if not self.check_user_allowed(user_data["login"]):
            raise UserNotAllowed(user_data["login"])

        jwt_token = self.token_handler.create_token(user_data["login"])

        return jwt_token

    @staticmethod
    def _get_id_and_secret():
        """
        Gets id and secret from GitHub.

        :return: Tuple of id and secret
        """

        load_dotenv()
        try:
            client_id = os.environ["CLIENT_ID"]
            client_secret = os.environ["CLIENT_SECRET"]
        except KeyError:
            raise KeyError(
                "CLIENT_ID and CLIENT_SECRET not found in .env file"
            ) from KeyError

        if client_id == "" or client_secret == "":
            raise KeyError("CLIENT_ID or CLIENT_SECRET is empty")

        return client_id, client_secret

    @staticmethod
    def get_method():
        """
        Gets login method.

        :return: str
        """

        return "github"

    @staticmethod
    def get_additional_data():
        """
        Gets additional data for login.

        :return: None
        """

        load_dotenv()
        try:
            client_id = os.environ["CLIENT_ID"]
        except KeyError:
            raise KeyError("CLIENT_ID not found in .env file") from KeyError

        if client_id == "":
            raise KeyError("CLIENT_ID is empty")

        return {"clientId": client_id}

    def authorize(self):
        """
        Authorizes GitHub login.

        :return: None
        """

        try:
            jwt_token = self.login()
            resp = make_response(redirect("/"))
            resp.set_cookie("token", jwt_token)
        except UserNotAllowed as e:
            self.logger.error(e)
            resp = make_response(redirect("/"))
            resp.status_code = 401
            resp.set_cookie("error", str(e))

        return resp

    def _get_access_token(self, request_token: str):
        """
        Gets access token from GitHub.

        :param request_token:
        :return: str
        """

        client_id, client_secret = self._get_id_and_secret()

        auth_data = {
            "code": request_token,
            "client_id": client_id,
            "client_secret": client_secret,
            "redirect_uri": "http://127.0.0.1:5000/github/callback",
        }
        resp = requests.post(
            "https://github.com/login/oauth/access_token", data=auth_data, timeout=10
        )
        token = resp.text.split("access_token=")[1]
        token = token.split("&")[0]

        return token

    @staticmethod
    def _get_user_data(access_token: str) -> dict:
        """
        Gets user data from GitHub.

        :param access_token: Access token
        :return: dict
        """
        if not access_token:
            raise ValueError("The request token has to be supplied!")
        if not isinstance(access_token, str):
            raise ValueError("The request token has to be a string!")

        user = requests.get(
            "https://api.github.com/user",
            headers={"Authorization": "token " + access_token},
            timeout=10,
        )

        return user.json()



from nerfw.helpers.logger import LoggerBase
from nerfw.helpers.breaker import Breaker
from nerfw.helpers.img_handler import ImageHandler


class Deconstructor(LoggerBase):
    """Class for scene deconstruction"""

    def __init__(self):
        super().__init__()
        self.image_handler = ImageHandler()

    def deconstruct(self, breaker: Breaker):
        """
        Deconstructs Breaker exception into json

        :param breaker: Exception to deconstruct
        :return: Dict
        """

        self.logger.debug(f"Deconstructing: {breaker}")
        scene_dict = {
            "background": breaker.scene.background,
            "characters": [
                character.to_dict() for character in breaker.scene.characters_to_show
            ],
            "text": breaker.line,
            "name": breaker.scene.name,
            "color": breaker.scene.color,
            "choice": {}
            if breaker.scene.choice is None
            else breaker.scene.choice.to_dict(),
            "stringInput": {}
            if breaker.scene.string_input is None
            else breaker.scene.string_input.to_dict(),
            "audio": breaker.scene.audio,
        }

        return scene_dict



from nerfw.server.token_handler import TokenHandler


class FlaskAppWrapper:
    """
    Wrapper for a flask app
    Configures the server
    """

    def __init__(self, app, **configs):
        self.app = app
        self.config = {}
        self.configs(**configs)
        self.app.secret_key = TokenHandler().get_key()

    def configs(self, **configs):
        """
        Adds configs to the app

        :param configs: Configurations for the server
        :return: None
        """
        for config, value in configs:
            self.app.config[config.upper()] = value

    def add_endpoint(
        self,
        endpoint=None,
        endpoint_name=None,
        handler=None,
        methods=None,
        *args,
        **kwargs
    ):
        """
        Adds server url endpoints

        :param endpoint: url
        :param endpoint_name: name for the endpoint
        :param handler: Function which is called
        :param methods: Web methods
        :param args: arguments which are passed to the function
        :param kwargs: positional args which are passed to function
        :return: None
        """
        if methods is None:
            methods = ["GET"]
        self.app.add_url_rule(
            endpoint, endpoint_name, handler, methods=methods, *args, **kwargs
        )

    def run(self, **kwargs):
        """
        Runs a server

        :param kwargs: Positional arguments passed to the app
        :return: None
        """
        self.app.run(**kwargs)



class UserNotAllowed(Exception):
    """
    Exception for user not allowed
    """
    def __init__(self, user_name: str):
        self.message = f"User {user_name} is not allowed"
        super().__init__(self.message)



class WrongNamingFormatError(Exception):
    """Raised when character can't be found"""

    def __init__(self, name):
        message = f"Wrong formatting for '{name}'. Name's should not have any spaces"
        super().__init__(message)



from nerfw.game.audio import Audio
from nerfw.game.character import Character
from nerfw.game.choice import Choice
from nerfw.game.string_input import StringInput
from nerfw.game.unlock_scene import UnlockScene
from nerfw.helpers.recorder import Recorder
from nerfw.game.scene import Scene
from nerfw.helpers.logger import LoggerBase


class Game(LoggerBase):
    """Class to create a login_register"""

    def __init__(self, *args, **kwargs):
        _ = args
        username = kwargs.get("username")
        last_line = kwargs.get("last_line")
        super().__init__()
        self._username = username
        self.scene = Scene()
        self.unlocker = UnlockScene()
        self._recorder = Recorder(last_line, self.scene)
        self._audio = Audio(self._recorder)
        self._choice_count = 0
        self._string_count = 0

    def create_character(
        self, name: str, img: str, color: (int, int, int), pos: (int, int)
    ):
        """
        Generates a character_files class

        :param name: Name of the character_files
        :param img: Path to image file
        :param color: Color in rgb from 0 to 255
        :param pos: Position of the character on screen in percent
        :return: Character class
        """

        character = Character(self._recorder, name, img, color, pos)
        return character

    def set_background(self, img_path: str):
        """
        Sets background to display

        :param img_path: Path to image file
        :return: None
        """

        self.logger.debug(f"Setting background to: {img_path}")
        self.scene.set_background(img_path)

    def choice(self, choices: list):
        """
        Makes a choice

        :param choices: List of options
        :return: returns a list with booleans representing answers
        """

        choice = Choice(
            self._recorder, choices, self._recorder.previous, self._choice_count
        )
        self._recorder.scene.add_choice_to_scene(choice)
        self._choice_count += 1
        answers = choice.make()

        return answers

    def string_input(self, string_name: str):
        """
        Allows string input

        :param string_name: Name of the variable to display
        :return: Answer from user
        """

        string_input = StringInput(
            self._recorder, self._recorder.previous, string_name, self._string_count
        )
        self._recorder.scene.add_string_input(string_input)
        self._string_count += 1
        answer = string_input.make()

        return answer

    def play_audio(self, filename: str, repeat: bool = False):
        """
        Plays audio

        :param filename: File to play
        :param repeat: Whether to repeat audio or not
        :return: None
        """

        self._audio.play(filename, repeat)

    def stop_audio(self, filename: str):
        """
        Stops audio

        :param filename: File to stop
        :return: None
        """

        self._audio.stop_playing(filename)

    def unlock_scene(self, img_path: str, label: str, category: str = None):
        """
        Unlocks scene

        :param img_path: Path to image file
        :param category: Category of the scene
        :param label: Label of the scene. Should be unique
        :return: None
        """

        self.unlocker.unlock(img_path, self._username, label, category)
        self.logger.info(f"{self._username} Unlocked scene: {img_path}")



import os
import logging


class LoggerBase:
    """ Class for log handling """

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)

    @staticmethod
    def setup_logger():
        """
        Sets up logging

        :return: None
        """

        if not os.path.isdir('logs'):
            os.makedirs('logs')

        fmt = '%(asctime)s [%(name)s]: %(levelname)s: %(message)s'

        stdout_handler = logging.StreamHandler()
        stdout_handler.setLevel(logging.DEBUG)

        file_handler = logging.FileHandler("logs/logs.log", 'a')
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(logging.Formatter(fmt))

        logging.basicConfig(
            level=logging.DEBUG,
            handlers=[
                stdout_handler,
                file_handler
            ]
        )



import os.path

from nerfw.helpers.img_handler import ImageHandler

from nerfw.game.animation import Animations
from nerfw.helpers.logger import LoggerBase


class Character(LoggerBase):
    """Class for character_files"""

    # pylint: disable=too-many-arguments
    def __init__(self, recorder=None, name="Test name", img_path="/", color=(0, 0, 0), pos=(0, 0)):
        super().__init__()
        self.recorder = recorder
        self.name = name
        self.img = img_path
        self.img_handler = ImageHandler()
        self.color = color
        self.animation = Animations(pos[0], pos[1])
        self.img_scale = {
            "height": 100,
            "width": 100
        }

    def to_dict(self):
        """
        Converts to dict

        :return: dict
        """

        output = {
            "name": self.name,
            "color": self.color,
            "img": self.img_handler.convert_to_base64(self.img),
            "scale": self.img_scale,
            "css": self.animation.css
        }

        return output

    def say(self, text="Sample text"):
        """
        Character will display text

        :param text: String to display
        :return: None
        """

        self.logger.info(f"Adding to script: {text}")
        self.recorder.check(text, self.name, self.color)

    def hide(self):
        """
        Removes character img from scene

        :return: None
        """

        self.logger.info("Hiding")
        self.recorder.scene.remove_character_from_scene(self)

    def show(self):
        """
        Shows character on display

        :return: None
        """

        self.recorder.scene.add_character_to_scene(self)

    def scale(self, height: int, width: int):
        """
        Scales character

        :param height: Height of the character image in px
        :param width: Width of the character image in px
        :return: None
        """

        self.img_scale = {
            "height": f"{height}px",
            "width": f"{width}px"
        }

    def set_image(self, img_path: str):
        """
        Sets character image

        :param img_path: Path to image
        :return: None
        """

        if not os.path.exists(img_path):
            self.logger.error(f"Path {img_path} is a directory")
            raise FileNotFoundError(f"Path {img_path} is a directory")

        self.logger.debug(f"Setting img to {img_path}")
        self.img = img_path



import json
import datetime
import logging
from json import loads

from authlib.integrations.flask_client import OAuth
from flask import Flask, request, jsonify, render_template, make_response, redirect
from nerfw.helpers.deconstructor import Deconstructor

from nerfw.helpers.breaker import Breaker
from nerfw.helpers.errors.password_mismatch import PasswordsMismatch
from nerfw.helpers.errors.user_already_registered import UserAlreadyRegistered
from nerfw.helpers.errors.user_doesnt_exist import UserDoesntExist
from nerfw.helpers.errors.user_not_allowed import UserNotAllowed
from nerfw.helpers.input_handler import InputHandler
from nerfw.server.gallery import Gallery
from nerfw.server.login_github import LoginGithub
from nerfw.server.login_default import LoginDefault
from nerfw.server.login_patreon import LoginPatreon
from nerfw.server.require_token import require_token
from nerfw.server.saves_handler import SavesHandler
from nerfw.server.token_handler import TokenHandler
from nerfw.server.wrapper import FlaskAppWrapper


class Server:
    """Server class"""

    def __init__(self, login_method: str, name: str):
        flask_app = Flask(
            __name__,
            static_url_path="/",
            static_folder="build",
            template_folder="build",
        )
        self.logger = logging.getLogger(self.__class__.__name__)
        flask_app.errorhandler(404)(lambda e: render_template("index.html"))
        self.app = FlaskAppWrapper(flask_app)
        self.input = InputHandler("", "")
        self.deconstructor = Deconstructor()
        self.saves_handler = SavesHandler()
        self.login_handler = None
        self._set_login_method(login_method)
        self.token_handler = TokenHandler()
        self.gallery_handler = Gallery()
        self.script = None
        self.name = name
        self.background = {"type": "background", "data": "#282a36"}

    def set_background(self, background: dict):
        """
        Sets background

        :param background: Background
        :return: None
        """

        self.background = background

    def _set_login_method(self, method: str):
        """
        Sets login method

        :param method: Login method. Can be "Default", "GitHub"
        :return: None
        """

        if method == "default":
            self.login_handler = LoginDefault()
        elif method == "github":
            oauth = OAuth(self.app)
            self.login_handler = LoginGithub(oauth)
        elif method == "patreon":
            self.login_handler = LoginPatreon()

    def home(self, path: str):
        """
        Renders home url

        :param path: Path
        :return: Rendered template from html
        """

        resp = make_response(render_template("index.html", name_game=self.name))
        if path:
            self.logger.info(f"Path requested: {path}")
        self.input.reset()

        return resp

    def api_authorize(self):
        """
        Login page

        :return: Login page
        """

        if request.method == "GET":
            resp_data = {
                "method": self.login_handler.get_method(),
                "additionalData": self.login_handler.get_additional_data(),
            }
            resp = make_response(resp_data)
            return resp

        data = request.get_json()
        try:
            self.login_handler.login(data)
            resp = make_response()
            token = self.token_handler.create_token(data["Login"])
            expire_date = datetime.datetime.now()
            expire_date = expire_date + datetime.timedelta(days=7)
            resp.set_cookie("token", token, httponly=False, expires=expire_date)
            resp.set_cookie("line", self.input.get_current_line())
            resp.set_cookie("prev_line", self.input.get_prev_line())
        except UserDoesntExist:
            resp = make_response()
            resp.status = 404
        except UserNotAllowed:
            resp = make_response()
            resp.status = 401

        return resp

    def api_register(self):
        """
        Registers a user

        :return: Register page
        """

        data = request.get_json()
        try:
            self.login_handler.register(data)
            resp = make_response()
        except UserAlreadyRegistered:
            resp = make_response()
            resp.status = 409
        except PasswordsMismatch:
            resp = make_response()
            resp.status = 401
            resp.set_cookie("error", "Password mismatch")

        return resp

    @require_token
    def api_backward(self):
        """
        Returns to previous slide

        :return: None
        """

        line = request.cookies.get("prev_line")

        if line is None:
            line = {"line": "", "back": False, "choices": {}, "stringInput": {}}
            line = json.dumps(line)

        resp = make_response()
        resp.set_cookie("line", line)
        self.input.cookie["lines"]["current"] = json.loads(line)
        self.input.cookie["lines"]["previous"] = json.loads(line)

        return resp

    @require_token
    def api_forward(self):
        """
        Progresses the game forward function

        :return: None
        """

        line = request.cookies.get("line")
        username = self.token_handler.unlock_token(request.cookies.get("token"))[
            "login"
        ]
        if line is None:
            line = {"line": "", "back": False, "choices": {}, "stringInput": {}}
        else:
            line = loads(line)

        try:
            answer = request.get_json()["choice"]
            choice_id = request.get_json()["id"]
            line["choices"][choice_id] = answer
        except KeyError:
            pass

        try:
            string_input = request.get_json()["stringInput"]
            input_id = request.get_json()["id"]
            line["stringInput"][str(input_id)] = string_input
        except KeyError:
            pass

        self.input.set_choices(line["choices"])
        self.input.set_string_input(line["stringInput"])
        line = json.dumps(line)

        try:
            self.script(last_line=line, username=username)
            output = redirect("/")
            text = ""
        except Breaker as br:
            output = self.deconstructor.deconstruct(br)
            text = br.line

        resp = make_response(output)
        self.input.set_line(text)
        resp.set_cookie("line", self.input.get_current_line())
        resp.set_cookie("prev_line", self.input.get_prev_line())

        return resp

    @require_token
    def api_get_saves(self):
        """
        Saves menu

        :return: None
        """

        if request.method == "GET":
            login = request.cookies.get("token")
            login = self.token_handler.unlock_token(login)["login"]
            saves = self.saves_handler.get_all_saves(login)

            output = self.render_save(saves, login)

            resp = make_response(output)
            return resp

        data = request.get_json()["data"]
        data = json.loads(data)
        resp = make_response()
        self.input.cookie["lines"]["previous"] = json.loads(data["prev_line"])
        self.input.set_line(json.loads(data["line"])["line"])
        self.input.cookie["lines"]["current"]["back"] = json.loads(data["line"])["back"]
        self.input.cookie["lines"]["current"]["choices"] = json.loads(data["line"])[
            "choices"
        ]

        return resp

    def render_save(self, saves, username):
        """
        Renders scene object for each save

        :param saves: Saves data to render
        :param username: Username
        :return: dict
        """

        output = []

        for date, save in saves:
            line = loads(save)
            line = line["line"]
            try:
                self.script(last_line=line, username=username)
                data = {}
            except Breaker as br:
                data = self.deconstructor.deconstruct(br)
            output.append({"date": date, "data": data, "save": save})

        return output

    @require_token
    def api_create_save(self):
        """
        Creates a save entry in db

        :return: None
        """

        data = request.cookies.to_dict()
        login = self.token_handler.unlock_token(data["token"])["login"]
        self.saves_handler.create_save(login, data)
        resp = make_response()
        return resp

    @require_token
    def api_get_gallery(self):
        """
        Gallery endpoint

        :return: None
        """

        login = request.cookies.get("token")
        login = self.token_handler.unlock_token(login)["login"]
        images = self.gallery_handler.get_images(login)
        images = [
            {"image": img, "category": category, "label": label}
            for img, category, label in images
        ]

        resp = make_response(jsonify(images))
        return resp

    @require_token
    def api_get_background(self):
        """
        Gets background image

        :return: None
        """

        resp = make_response(jsonify(background=self.background))
        return resp

    def run(self, script, debug=False, port=5000):
        """
        Runs a server

        :param debug: Sets debug mode
        :param script: Script to be run
        :param port: Port to be used
        :return: None
        """

        self.script = script
        try:
            self.script(last_line=self.input.get_current_line(), username="")
        except Breaker:
            pass

        self.app.add_endpoint(
            "/", "home", self.home, methods=["GET"], defaults={"path": ""}
        )
        self.app.add_endpoint("/<path:path>", "home", self.home, methods=["GET"])
        self.app.add_endpoint(
            "/api/forward", "forward", self.api_forward, methods=["POST"]
        )
        self.app.add_endpoint(
            "/api/backward", "backward", self.api_backward, methods=["POST"]
        )
        self.app.add_endpoint(
            "/api/save", "save", self.api_create_save, methods=["POST"]
        )
        self.app.add_endpoint(
            "/api/get_saves", "get_saves", self.api_get_saves, methods=["GET", "POST"]
        )
        self.app.add_endpoint(
            "/api/authorize", "authorize", self.api_authorize, methods=["POST", "GET"]
        )
        self.app.add_endpoint(
            "/api/register", "register", self.api_register, methods=["POST"]
        )
        self.app.add_endpoint(
            "/api/gallery", "gallery", self.api_get_gallery, methods=["GET"]
        )
        self.app.add_endpoint(
            "/api/get_background",
            "get_background",
            self.api_get_background,
            methods=["GET"],
        )

        if self.login_handler.get_method() == "github":
            self.app.add_endpoint(
                "/github/callback",
                "github_authorized",
                self.login_handler.authorize,
                methods=["GET"],
            )
        elif self.login_handler.get_method() == "patreon":
            self.app.add_endpoint(
                "/patreon/callback",
                "patreon_authorized",
                self.login_handler.authorize,
                methods=["GET"],
            )

        self.app.run(host="0.0.0.0", debug=debug, port=port)



from json import loads
from nerfw.game.scene import Scene
from nerfw.helpers.logger import LoggerBase
from nerfw.helpers.breaker import Breaker


class Recorder(LoggerBase):
    """Class to track progress"""

    def __init__(self, last_line: str, scene: Scene):
        super().__init__()
        self.previous = loads(last_line)
        self.scene = scene
        self.report = False

    def check(self, line: str, name="", color=(0, 0, 0)):
        """
        Checks if line is already encountered

        :param line: Checks the current line
        :param name: Name to display
        :param color: Color of the name to display
        :return: None
        """

        self.scene.name = name
        self.scene.color = color

        if self.report or (self.previous["line"] == ""):
            self.report = False
            raise Breaker(line, self.scene)

        if line == self.previous["line"] and not self.previous["back"]:
            self.report = True

        if line == self.previous["line"] and self.previous["back"]:
            raise Breaker(line, self.scene)



import sqlite3

from nerfw.helpers.logger import LoggerBase


class DbHandler(LoggerBase):
    """
    Handler for database
    """

    def __init__(self):
        super().__init__()
        self.path = "login_data.db"
        self.connection = sqlite3.connect(self.path, check_same_thread=False)
        try:
            self.setup_db()
        except sqlite3.OperationalError:
            pass

    def setup_db(self):
        """
        Sets up databases

        :return:
        """

        with self.connection:
            cur = self.connection.cursor()
            cur.execute(
                "CREATE TABLE IF NOT EXISTS saves (login TEXT, date TEXT, data TEXT);"
            )
            cur.execute(
                "CREATE TABLE IF NOT EXISTS credentials"
                "(login TEXT, password TEXT, salt TEXT, UNIQUE(login));"
            )
            cur.execute(
                "CREATE TABLE IF NOT EXISTS unlockedScenes "
                "(sceneId TEXT, user TEXT, label TEXT, image TEXT, "
                "category TEXT, "
                "date DATETIME DEFAULT CURRENT_TIMESTAMP, "
                "UNIQUE(sceneId));"
            )

    def execute(self, query: str, values=None):
        """
        Executes sql query

        :param query: Query to execute
        :param values: Values if have
        :return: list
        """

        with self.connection:
            cur = self.connection.cursor()
            if values is None:
                cur.execute(query)
            else:
                cur.execute(query, values)

        return cur.fetchall()



class UnsupportedLoginMethodError(Exception):
    """
    Exception raised when a login method is not supported.
    """
    def __init__(self, method: str):
        message = f"Unsupported login method: {method}"
        super().__init__(message)



import sqlite3

from nerfw.helpers.db_handler import DbHandler
from nerfw.helpers.img_handler import ImageHandler
from nerfw.helpers.logger import LoggerBase


class UnlockScene(LoggerBase):
    """
    Class to unlock scenes
    """

    def __init__(self):
        super().__init__()
        self.db = DbHandler()
        self.img_handler = ImageHandler()

    def unlock(self, image: str, username: str, label: str, category: str):
        """
        Unlocks a scene

        :param image: Path to image
        :param username: Username
        :param label: Label for the image
        :param category: Category of the image
        :return: None
        """

        self.logger.debug(f"Unlocking scene: {image}")
        binary = self.img_handler.convert_to_base64(image)
        query = ("INSERT INTO unlockedScenes (sceneId, user, label, image, category) "
                 "VALUES (?, ?, ?, ?, ?)")
        values = [(username + label), username, label, binary, category]

        try:
            self.db.execute(query, values)
        except sqlite3.IntegrityError:
            self.logger.debug(f" {username} has already unlocked {image}")



class UserAlreadyRegistered(Exception):
    """Raised when User is already exists"""

    def __init__(self, login: str):
        message = f"User '{login}' already exists"
        super().__init__(message)



from nerfw.helpers.logger import LoggerBase


class LoginBaseClass(LoggerBase):
    """
    Base class for login methods.
    """

    def __init__(self):
        super().__init__()
        self.allowed_users = []

    def check_user_allowed(self, login: str):
        """
        Checks if user is allowed

        :param login: User login
        :return: bool
        """

        return login in self.allowed_users if self.allowed_users else True

    @staticmethod
    def get_method():
        """
        Returns method name.

        :return: str
        """

        raise NotImplementedError

    @staticmethod
    def get_additional_data():
        """
        Gets additional data for login

        :return: None
        """

        return None

    def authorize(self):
        """
        Checks if user is authorized

        :return: None
        """

        return None

    def set_list_of_allowed_users(self, users: list):
        """
        Sets list of allowed users

        :param users: List of users
        :return: None
        """

        self.allowed_users = users



import json

from nerfw.helpers.logger import LoggerBase
from nerfw.helpers.db_handler import DbHandler


class SavesHandler(LoggerBase):
    """
    Class for handling saves files
    """

    def __init__(self):
        super().__init__()
        self.db = DbHandler()

    def create_save(self, login: str, data: dict):
        """
        Creates an entry in database

        :param login: User login
        :param data: Save data
        :return: None
        """

        self.logger.debug(f"Creating save for: {login}")
        sql = "INSERT INTO saves(login, date, data) VALUES(?, DATETIME('now','localtime'), ?)"
        data = {
            "line": data["line"],
            "prev_line": data["prev_line"]
        }
        text = json.dumps(data)
        values = [login, text]

        self.db.execute(sql, values)

    def get_all_saves(self, login: str):
        """
        Gets all save files for Login

        :param login: user login
        :return: list of saves
        """

        sql = "SELECT date, data FROM saves WHERE login = ?"
        values = [login]

        self.logger.debug(f"Requesting all saves for: {login}")
        result = self.db.execute(sql, values)

        return result



class CharacterNotFoundError(Exception):
    """Raised when character can't be found"""

    def __init__(self, character):
        message = f"Can't find character {character}"
        super().__init__(message)



from functools import wraps

from flask import request, redirect

from nerfw.server.token_handler import TokenHandler


def require_token(func):
    """
    Decorator to check if token is available

    :param func: Function to check
    :return: Wrapper
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        token = request.cookies.get("token")
        handler = TokenHandler()
        answer = handler.verify_token(token)
        if not answer:
            response = redirect("/login")
            response.set_cookie("token", "", expires=0)
            return response
        return func(*args, *kwargs)

    return wrapper



from json import dumps
from nerfw.helpers.logger import LoggerBase


class InputHandler(LoggerBase):
    """Class responsible for input handling"""

    def __init__(self, current_line: str, prev_line: str):
        super().__init__()
        self.cookie = {
            "lines": {
                "current": {
                    "line": current_line,
                    "back": False,
                    "choices": {},
                    "stringInput": {}
                },
                "previous": {
                    "line": prev_line,
                    "back": False,
                    "choices": {},
                    "stringInput": {}
                }
            }
        }

    def set_choices(self, choices: dict):
        """
        Sets choices in current line

        :param choices: Dict
        :return: None
        """

        self.cookie["lines"]["current"]["choices"] = choices

    def set_string_input(self, string_input: dict):
        """
        Sets string input in current line

        :param string_input: Dict
        :return: None
        """

        self.cookie["lines"]["current"]["stringInput"] = string_input

    def get_current_line(self):
        """
        Function to set input params when forward

        :return: JSON string
        """

        output = self.cookie["lines"]["current"]

        return dumps(output)

    def get_prev_line(self):
        """
        Function to set input params when backwards

        :return: JSON string
        """
        output = self.cookie["lines"]["previous"]

        return dumps(output)

    def set_line(self, line: str):
        """
        Sets new line as current

        :param line: Str
        :return: None
        """

        self.cookie["lines"]["previous"] = self.cookie["lines"]["current"].copy()
        self.cookie["lines"]["current"]["line"] = line

    def reset(self):
        """
        Resets to default

        :return: None
        """

        self.cookie = {
            "lines": {
                "current": {
                    "line": "",
                    "back": False,
                    "choices": {},
                    "stringInput": {}
                },
                "previous": {
                    "line": "",
                    "back": False,
                    "choices": {},
                    "stringInput": {}
                }
            }
        }



import os

import patreon
import requests
from dotenv import load_dotenv
from flask import make_response, redirect, request

from nerfw.helpers.errors.user_not_allowed import UserNotAllowed
from nerfw.helpers.login_base import LoginBaseClass
from nerfw.server.token_handler import TokenHandler


class LoginPatreon(LoginBaseClass):
    """
    Login to Patreon.
    """

    def __init__(self):
        super().__init__()
        self.token_handler = TokenHandler()

    @staticmethod
    def _get_id_and_secret():
        """
        Gets id and secret from GitHub.

        :return: Tuple of id and secret
        """

        load_dotenv()
        try:
            client_id = os.environ["CLIENT_ID"]
            client_secret = os.environ["CLIENT_SECRET"]
        except KeyError:
            raise KeyError(
                "CLIENT_ID and CLIENT_SECRET not found in .env file"
            ) from KeyError

        if client_id == "" or client_secret == "":
            raise KeyError("CLIENT_ID or CLIENT_SECRET is empty")

        return client_id, client_secret

    def login(self):
        """
        Login to Patreon.

        :return: JWT token
        """

        self.logger.info("Authorizing Patreon login...")
        client_id, client_secret = self._get_id_and_secret()
        oauth_client = patreon.OAuth(client_id, client_secret)
        tokens = oauth_client.get_tokens(
            request.args.get("code"), "http://127.0.0.1:5000/patreon/callback"
        )
        access_token = tokens["access_token"]
        user_response = self.get_user_data(access_token)
        campaign = self.get_campaign_id()

        if not self.check_user_is_patron(user_response, campaign):
            raise UserNotAllowed(
                user_response["included"][0]["attributes"]["full_name"]
            )

        jwt_token = self.token_handler.create_token(
            user_response["included"][0]["attributes"]["full_name"]
        )

        return jwt_token

    @staticmethod
    def get_creator_token():
        """
        Gets creator token.

        :return: str
        """

        load_dotenv()
        try:
            creator_token = os.environ["CREATOR_TOKEN"]
        except KeyError:
            raise KeyError("CREATOR_TOKEN not found in .env file") from KeyError

        if creator_token == "":
            raise KeyError("CREATOR_TOKEN is empty")

        return creator_token

    @staticmethod
    def get_method():
        """
        Gets login method.

        :return: str
        """

        return "patreon"

    @staticmethod
    def get_additional_data():
        """
        Gets additional data for login.

        :return: None
        """

        load_dotenv()
        try:
            client_id = os.environ["CLIENT_ID"]
        except KeyError:
            raise KeyError("CLIENT_ID not found in .env file") from KeyError

        if client_id == "":
            raise KeyError("CLIENT_ID is empty")

        redirect_url = "http://127.0.0.1:5000/patreon/callback"

        return {"clientId": client_id, "redirectUrl": redirect_url}

    def authorize(self):
        """
        Authorizes GitHub login.

        :return: None
        """

        try:
            jwt_token = self.login()
            resp = make_response(redirect("/"))
            resp.set_cookie("token", jwt_token)
        except UserNotAllowed as e:
            self.logger.error(e)
            resp = make_response(redirect("/"))
            resp.status_code = 401
            resp.set_cookie("error", str(e))

        return resp

    @staticmethod
    def get_user_data(access_token: str):
        """
        Gets user data.

        :param access_token: Access token
        :return:
        """

        url = (
            "https://www.patreon.com/api/oauth2/v2/identity"
            "?include=memberships.campaign&fields%5Bmember%5D=patron_status,full_name"
        )

        headers = {"Authorization": f"Bearer {access_token}"}
        response = requests.get(
            url,
            headers=headers,
            timeout=10,
        )

        return response.json()

    def get_campaign_id(self):
        """
        Gets current patrons.

        :return: Campaign id
        """

        creator_token = self.get_creator_token()
        headers = {"Authorization": f"Bearer {creator_token}"}
        response = requests.get(
            "https://patreon.com/api/oauth2/v2/campaigns?fields%5Bcampaign%5D=creation_name",
            headers=headers,
            timeout=10,
        )

        data = response.json()["data"]
        id_list = [campaign["id"] for campaign in data if campaign]

        return id_list[0]

    @staticmethod
    def check_user_is_patron(user_data: dict, campaign_id: str):
        """
        Checks if user is a patron.

        :param user_data: User data
        :param campaign_id: Campaign id
        :return: bool
        """

        for obj in user_data["included"]:
            try:
                if (
                    obj["relationships"]["campaign"]["data"]["id"] == campaign_id
                    and obj["attributes"]["patron_status"] == "active_patron"
                ):
                    return True
            except KeyError:
                pass

        return False



from nerfw.game.character import Character
from nerfw.game.string_input import StringInput
from nerfw.helpers.errors.character_not_found import CharacterNotFoundError
from nerfw.helpers.img_handler import ImageHandler
from nerfw.helpers.logger import LoggerBase


class Scene(LoggerBase):
    """Class which creates a scene description"""

    def __init__(self):
        super().__init__()
        self.background = None
        self.characters_to_show = []
        self.choice = None
        self.string_input = None
        self.audio = []
        self.name = ""
        self.color = ()
        self.img_handler = ImageHandler()

    def set_background(self, img_file: str):
        """
        Sets background

        :param img_file:
        :return: None
        """

        self.logger.info(f"Setting Background to {img_file}")
        self.background = self.img_handler.convert_to_base64(img_file)

    def add_character_to_scene(self, character: Character):
        """
        Adds character to the displayed scene

        :param character: Character class to be shown
        :return: None
        """

        if character not in self.characters_to_show:
            self.logger.info(f"Adding to display: {character.name}")
            self.characters_to_show.append(character)

    def remove_character_from_scene(self, character: Character):
        """
        Removes character from display

        :param character: Character class to be removed
        :return: None
        """

        if character in self.characters_to_show:
            self.characters_to_show.remove(character)
        else:
            raise CharacterNotFoundError(character)

    def add_choice_to_scene(self, choice):
        """
        Adds choice to be displayed

        :param choice: Choice object
        :return: None
        """

        self.choice = choice
        self.logger.debug(f"Setting choice to: {choice}")

    def remove_choice_from_scene(self):
        """
        Removes choice from object

        :return: None
        """

        self.logger.debug("Removing choice from scene object")
        self.choice = None

    def add_audio(self, filename: str, repeatable: bool = False):
        """
        Adds audio to self

        :param filename: Audio file name
        :param repeatable: If audio should be repeatable
        :return: None
        """

        self.logger.debug(f"Added {filename}")
        file_info = {"name": filename, "repeatable": repeatable}
        self.audio.append(file_info)

    def remove_audio(self, filename: str):
        """
        Removes audio from self

        :param filename: Audio file name
        :return: None
        """

        self.logger.debug(f"Removed {filename}")
        self.audio = [file for file in self.audio if file["name"] != filename]

    def add_string_input(self, string_input: StringInput):
        """
        Adds string input to self

        :param string_input: String input object
        :return: None
        """

        self.logger.debug(f"Added {string_input}")
        self.string_input = string_input

    def remove_string_input(self):
        """
        Removes string input from self

        :return: None
        """

        self.logger.debug("Removed string input")
        self.string_input = None



import base64
import io
import zlib
from PIL import Image

from nerfw.helpers.logger import LoggerBase


class ImageHandler(LoggerBase):
    """Class for handling images"""

    def __init__(self):
        super().__init__()
        self.logger.debug("Initialized")

    def _compress(self, img_b64: bytes):
        """
        Compresses img

        :param img_b64: Base64 encoded img
        :return: Compressed img bytes
        """

        self.logger.debug("Compressing...")
        output = zlib.compress(img_b64, level=9)

        return output

    def convert_to_base64(self, img_path: str):
        """
        Converts img and returns base64

        :param img_path: Path to img file
        :return: Base64 encoded img
        """

        self.logger.info(f"Converting {img_path} to base64")
        with open(img_path, "rb") as img_file:
            img = Image.open(img_file)
            buffer = io.BytesIO()
            img.save(buffer, format="WEBP")
            encoded_image = base64.b64encode(buffer.getvalue())

        compressed = self._compress(encoded_image)

        return base64.b64encode(compressed).decode("ascii")



from nerfw.helpers.logger import LoggerBase


class StringInput(LoggerBase):
    """
    Allows string input
    """

    def __init__(self, recorder=None, cookies=None, string=None, string_id=None):
        super().__init__()
        self.recorder = recorder
        self.string_name = string
        self.string_id = string_id
        self.answer = None
        self.cookies = cookies
        self.setup()

    def setup(self):
        """
        Sets up string from cookies

        :return: None
        """

        try:
            self.logger.debug(
                f"Found existing string: {self.cookies['stringInput'][str(self.string_id)]}"
            )
            self.answer = self.cookies["stringInput"][str(self.string_id)]
        except KeyError:
            self.answer = ""

    def make(self):
        """
        Displays or passes a string

        :return: string
        """

        self.recorder.check(f"{self.string_name}{self.string_id}")
        self.recorder.scene.remove_string_input()
        return self.answer

    def to_dict(self):
        """
        Converts obj to dict

        :return: dict
        """

        return {"id": self.string_id, "text": self.string_name}



from nerfw.helpers.db_handler import DbHandler
from nerfw.helpers.img_handler import ImageHandler
from nerfw.helpers.logger import LoggerBase


class Gallery(LoggerBase):
    """
    Class to handle gallery
    """

    def __init__(self):
        super().__init__()
        self.img_handler = ImageHandler()
        self.db = DbHandler()

    def get_images(self, username: str):
        """
        Gets images for the user

        :param username: Username
        :return: list
        """

        self.logger.debug(f"Getting images for {username}")
        query = "SELECT image, category, label FROM unlockedScenes WHERE user = ?"
        result = self.db.execute(query, [username])
        self.logger.debug(f"Images for {username}: {result}")
        return result



import hashlib
import random
import sqlite3
import string

from nerfw.helpers.errors.user_already_registered import UserAlreadyRegistered

from nerfw.helpers.db_handler import DbHandler

from nerfw.helpers.logger import LoggerBase


class PasswordManager(LoggerBase):
    """
    Class to manage passwords in db
    """

    def __init__(self, db: DbHandler):
        super().__init__()
        self.db = db

    def register(self, data: dict):
        """
        Registers a password in db

        :param data: Data from web
        :return: None
        """

        pwd = data["Password"]
        self.logger.debug(f"Hashing password for {data['Login']}")
        hashed, salt = self.encrypt(pwd)
        sql = "INSERT INTO credentials(login, salt, password) VALUES(?, ?, ?)"
        vals = [data["Login"], salt, hashed]
        try:
            self.db.execute(sql, vals)
        except sqlite3.IntegrityError as e:
            raise UserAlreadyRegistered(data["Login"]) from e

    @staticmethod
    def encrypt(password: str, salt=None):
        """
        Encrypts a password

        :param password: Provided password
        :param salt: Salt to be used
        :return: Hashed password
        """

        if salt is None:
            salt = "".join(random.choices(string.ascii_letters + string.digits, k=25))

        pwd = f"{password}{salt}"
        hashed = hashlib.sha256(pwd.encode()).hexdigest()

        return hashed, salt



from nerfw.helpers.errors.user_not_allowed import UserNotAllowed
from nerfw.helpers.db_handler import DbHandler
from nerfw.helpers.errors.password_mismatch import PasswordsMismatch
from nerfw.helpers.errors.user_doesnt_exist import UserDoesntExist
from nerfw.helpers.login_base import LoginBaseClass
from nerfw.server.password_manager import PasswordManager


class LoginDefault(LoginBaseClass):
    """
    Handles login process
    """

    def __init__(self):
        super().__init__()
        self.db = DbHandler()
        self.pwd_manager = PasswordManager(self.db)

    def login(self, data: dict):
        """
        Checks user data

        :param data: Login credentials
        :return: dict
        """

        if not self.check_user_allowed(data["Login"]):
            raise UserNotAllowed(data["Login"])

        sql = "SELECT password, salt FROM credentials WHERE login = ?"
        result = self.db.execute(sql, [data["Login"]])
        self.logger.debug(f"Found results: {result}")
        if len(result) != 1:
            raise UserDoesntExist(data["Login"])

        hashed, _ = self.pwd_manager.encrypt(data["Password"], result[0][1])
        if hashed != result[0][0]:
            raise PasswordsMismatch()

        return result[0]

    def register(self, data: dict):
        """
        Registers new user

        :param data: Login credentials
        :return: bool
        """

        self.logger.debug(f"Received: {data}")
        if data["Password"] != data["Repeat_password"]:
            raise PasswordsMismatch()
        self.pwd_manager.register(data)

    @staticmethod
    def get_method():
        """
        Gets login method

        :return: str
        """

        return "default"



from nerfw.game.scene import Scene


class Breaker(Exception):
    """Breaks the script execution and returns a line"""

    def __init__(self, line: str, scene: Scene):
        self.line = line
        self.scene = scene
        super().__init__(self.line)



import typing

from nerfw.helpers.errors.unsupported_login_method import UnsupportedLoginMethodError
from nerfw.helpers.img_handler import ImageHandler
from nerfw.helpers.logger import LoggerBase
from nerfw.server.server import Server


class NerFW(LoggerBase):
    """
    Main class for Ner framework
    """

    def __init__(self):
        super().__init__()
        self.logger.info("Initializing NerFW")
        self.login_method = "default"
        self.allowed_users = []
        self.image_handler = ImageHandler()
        self.name = "NerFW"
        self.background = {"type": "background", "data": "#282a36"}

    def run(self, script, debug=False):
        """
        Runs a game

        :param script: Script to run
        :param debug: Debug mode
        :return: None
        """

        server = Server(self.login_method, self.name)
        server.set_background(self.background)
        server.login_handler.set_list_of_allowed_users(self.allowed_users)
        self.logger.info("Launching NerFW")
        server.run(script, debug=debug)

    def set_login_method(self, method: str):
        """
        Sets login method

        :param method: Login method. Can be "Default", "GitHub"
        :return: None
        """

        if method.lower() not in ["default", "github", "patreon"]:
            raise UnsupportedLoginMethodError(method)

        self.logger.info(f"Setting login method to {method}")
        self.login_method = method

    def set_allowed_users(self, users: typing.Union[str, list]):
        """
        Sets list of allowed users

        :param users: path to file with list of users. Or list of users
        :return: None
        """

        if isinstance(users, list):
            self.allowed_users = users
            return

        with open(users) as f:
            users = f.readlines()
        self.allowed_users = users

    def set_name(self, name: str):
        """
        Sets name of the game

        :param name: Name of the game
        :return: None
        """

        self.logger.info(f"Setting name of the game to {name}")
        self.name = name

    def set_background(self, background: str):
        """
        Sets background

        :param background: Path to background
        :return: None
        """

        self.logger.info(f"Setting background to {background}")
        img = self.image_handler.convert_to_base64(background)

        self.background = {
            "type": "backgroundImage",
            "data": img,
        }



import datetime
import os
import secrets

import jwt

from nerfw.helpers.logger import LoggerBase


class TokenHandler(LoggerBase):
    """
    Class for handling login token
    """

    def __init__(self):
        super().__init__()
        if not os.path.exists("data"):
            self.logger.info("Creating a data folder")
            os.mkdir("data")

        if os.path.exists("data/key.ner"):
            self.key = self.get_key()
        else:
            self.generate_key()
            self.key = self.get_key()

    def get_key(self):
        """
        Gets a key from file

        :return: key
        """

        with open("data/key.ner", "r") as file:
            key = file.read()

        self.logger.info("Received a key")
        return key

    def generate_key(self):
        """
        Generates a secret key

        :return: None
        """

        key = secrets.token_hex()
        with open("data/key.ner", "w") as file:
            file.write(key)
        self.logger.info("Generated a key at data/key.ner")

    def create_token(self, user_name: str):
        """
        Creates a session token for user

        :param user_name: Name of user
        :return: token
        """

        payload = {"login": user_name, "created": str(datetime.datetime.now())}

        token = jwt.encode(payload=payload, key=self.key, algorithm='HS256')

        return token

    def verify_token(self, token: str):
        """
        Verifies a jwt

        :param token: Lwt str token
        :return: bool or dict
        """

        if token is None:
            return False

        try:
            decoded = jwt.decode(token, key=self.key, algorithms='HS256')
        except jwt.ExpiredSignatureError:
            self.logger.debug("Expired signature")
            return False
        except jwt.InvalidTokenError:
            self.logger.debug("Invalid token")
            return False
        return decoded

    def unlock_token(self, token: str):
        """
        Decodes token and return json obj

        :param token: JWS token
        :return: dict
        """

        decoded = jwt.decode(token, key=self.key, algorithms='HS256')

        return decoded



import os
import subprocess
from importlib.resources import files
from pathlib import Path

from nerfw.helpers.logger import LoggerBase


class Audio(LoggerBase):
    """
    Class for audio control
    """

    def __init__(self, recorder=None):
        super().__init__()
        self.recorder = recorder

    def play(self, filename: str, repeat: bool = False):
        """
        Adds audio to html to be shown

        :param filename: Audio file to play
        :param repeat: If audio should be repeatable
        :return: None
        """

        file = filename.split("/")[-1]

        dir_folder = files("nerfw.server.build")
        self.logger.debug(dir_folder)
        # noinspection PyTypeChecker
        path = Path(dir_folder.joinpath(file))

        if not os.path.exists(path):
            if os.name == "nt":
                arg = ["copy", filename, str(path)]
            else:
                arg = ["cp", filename, str(path)]
            subprocess.call(arg)

        self.logger.debug(f"Adding {filename} to scene")
        self.recorder.scene.add_audio(f"{file}", repeatable=repeat)

    def stop_playing(self, filename: str):
        """
        Stops audio from playing

        :param filename: Audio file to stop
        :return: None
        """

        filename = filename.split("/")[-1]
        self.logger.debug(f"Stopping {filename}")
        self.recorder.scene.remove_audio(filename)



class FunctionNotFound(Exception):
    """Raised when character can't be found"""

    def __init__(self, func):
        message = f"Can't find specified function '{repr(func)}'"
        super().__init__(message)



from nerfw.helpers.logger import LoggerBase


class Choice(LoggerBase):
    """
    Class for creating a choice menu
    """

    def __init__(self, recorder=None, choices=None, cookies=None, choice_id=None):
        super().__init__()
        self.id = str(choice_id)
        if choices is None:
            choices = []
        self.recorder = recorder
        self.choices = choices
        self.answers = None
        self.cookies = cookies
        self.setup()

    def to_dict(self):
        """
        Converts obj to dict

        :return: dict
        """

        output = {"options": []}
        for choice in self.choices:
            output['options'].append(
                {
                    "id": self.id,
                    "text": choice
                }
            )

        return output

    def setup(self):
        """
        Sets up choice from cookies

        :return: None
        """

        try:
            self.logger.debug(f"Found existing answers: {self.cookies['choices'][self.id]}")
            self.answers = []
            for choice in self.choices:
                if choice == self.cookies["choices"][self.id]:
                    self.answers.append(True)
                else:
                    self.answers.append(False)
        except KeyError:
            self.answers = []

    def make(self):
        """
        Displays or passes a choice

        :return: answers
        """

        self.recorder.check(str(self.choices))
        self.recorder.scene.remove_choice_from_scene()
        return self.answers



from nerfw.helpers.logger import LoggerBase


class Animations(LoggerBase):
    """
    Class for handling animations
    """

    def __init__(self, pos_x, pos_y):
        super().__init__()
        self.current_x = pos_x
        self.current_y = pos_y
        self.css = {}

    def move(self, move_x: int, move_y: int, duration: float or int):
        """
        Moves character

        :param move_x: Translation by X axis
        :param move_y: Translation by Y axis
        :param duration: Duration of animation in seconds
        :return:
        """

        self.logger.debug(f"Moving from {self.current_x} to {move_x} by X\n"
                          f"Moving from {self.current_y} to {move_y} by Y")

        css = {
            "position": "absolute",
            "--moveStartX": f"{self.current_x}%",
            "--moveEndX": f"{move_x}%",
            "--moveStartY": f"{self.current_y}%",
            "--moveEndY": f"{move_y}%",
            "animation": f"{duration}s ease-in-out 0s 1 slideIn",
            "top": f"{move_y}%",
            "left": f"{move_x}%",
            "overflow": "hidden",
        }

        self.current_x = move_x
        self.current_y = move_y

        self.css = css
